print("import: 'pept'")
import pept

print("import: 'pept.base'")
import pept.base

print("import: 'pept.cookbook'")
import pept.cookbook

print("import: 'pept.diagnostics'")
import pept.diagnostics

print("import: 'pept.scanners'")
import pept.scanners

print("import: 'pept.scanners.modular_camera'")
import pept.scanners.modular_camera

print("import: 'pept.scanners.parallel_screens'")
import pept.scanners.parallel_screens

print("import: 'pept.simulation'")
import pept.simulation

print("import: 'pept.tracking'")
import pept.tracking

print("import: 'pept.tracking.birmingham_method'")
import pept.tracking.birmingham_method

print("import: 'pept.tracking.peptml'")
import pept.tracking.peptml

print("import: 'pept.tracking.trajectory_separation'")
import pept.tracking.trajectory_separation

print("import: 'pept.utilities'")
import pept.utilities

print("import: 'pept.utilities.cutpoints'")
import pept.utilities.cutpoints

print("import: 'pept.utilities.misc'")
import pept.utilities.misc

print("import: 'pept.utilities.parallel'")
import pept.utilities.parallel

print("import: 'pept.utilities.traverse'")
import pept.utilities.traverse

print("import: 'pept.visualisation'")
import pept.visualisation

